#include <stdio.h>
#include <stdlib.h>

#include <error.h>
#include <dbgmem.h>
#include <hash.h>
#include <file.h>

void freeint(char *str, void *ptr)
{
  free(ptr);
}

void testhash(void)
{
  hashtable h;
  int i;
  int *x; 
  char str[100];
  int ok,y;

  printf("Testing hashtable module.\n");
  DbgMemStat();

  h=HashCreate(10); printf("HashCreate(10)\n");
  DbgMemStat();

  printf("Inserting 100 random strings\n");
  srand(42);
  for(i=0;i<100;i++) {
    x=malloc(sizeof(int));
    *x=rand();
    sprintf(str,"%d",*x);
    HashInsert(h,x,str);  
  }  
  
  printf("Checking the strings\n");
  srand(42);
  ok=1;
  for(i=0;i<100;i++) {
    sprintf(str,"%d",y=rand()); 
    x=HashLookup(h,str);  
    if (x==0) {
        printf("%s Not found\n",str);
        ok=0;
    } else {
        if (y!=*x) {
          printf("Couldn't find %s\n",str);
          ok=0;
        }
    }
  }  
  if (!ok) {
    printf("There is an error in the hashtable module\n");
  } else {
    printf("Checking was ok.\n");
  }

  HashApply(h,freeint);printf("Freeing strings\n");
  HashDestroy(h); printf("HashDestroy()\n");
  DbgMemStat();

}

void testlog(void)
{
  printf("Testing error module.\n");
  ErrSetLogName("mupp.err");
  ErrLog("Testar ErrLog");
  ErrWarning("Testar ErrWarning");
/*  ErrError("Testar ErrError");*/

  verify(printf("verify()\n"));
  assert(printf("assert()\n"));
/*  verify(printf("verify()\n")==2);
  assert(printf("assert()\n")==2);*/

}

void testmem(void)
{
  char *a,*b,*c;
  
  printf("testing dbgmem module.\n");
  a=calloc(1024,sizeof(char));printf("malloc(a)\n");
  b=calloc(1024,sizeof(char));printf("malloc(b)\n");
  c=calloc(1024,sizeof(char));printf("malloc(c)\n");
  DbgMemStat();
  a=realloc(a,1023);printf("realloc(a,1023)\n");
  DbgMemStat();
  b=realloc(b,1025);printf("realloc(b,1025)\n");
  DbgMemStat();
  free(b);printf("free(b)\n");
  DbgMemStat();
  free(a);printf("free(a)\n");
  DbgMemStat();
  free(c);printf("free(c)\n");
  DbgMemStat();
}


void createdatafile(void)
{
  char *str;

  printf("Testing datafile module:\n");

  printf("Creating data file 'test.dat'\n");
  if (FileNew()) 
    ErrError("New()");
  printf("Adding file 'include/zobject.h'\n");
  if (FileAdd("include/zobject.h"))
    ErrError("Can't add file.h");
  printf("Adding file 'common/main.c'\n");
  if (FileAdd("common/main.c"))
    ErrError("Can't add main.h");
  if (FileCreate("test.dat"))
    ErrError("Can't creat file");

  printf("opening 'test.dat'\n");
  if (FileOpen("test.dat"))
    ErrError("Can't open test.dat");
  printf("Listing files:\n");
  str=FileListFirst();
  while(str!=0) {
    printf("%s %ld bytes\n",str,FileSize(str));
    str=FileListNext();
  }
  printf("Showing file include/zobject.h:\n");
  str=FileLoad("include/zobject.h");
  if (str==0)
    ErrError("Couldn't load file");
  str[FileSize("include/zobjetc.h")-1]=0;
  printf("%s",str);
  FileClose();
  DbgMemStat();
}



void main(void)
{

  printf("Testing things\n");

  testmem();
  testhash();
  testlog();

  createdatafile();

}
