#include <arch.h>
#include <device.h>
#include <error.h>
#include <dbgmem.h>
#include <common.h>
#include "vector.h"
#include <object.h>
#include <zobject.h>
#include "render.h"
#include "zsort.h"
#include "fastmem.h"
#include "engine.h"

vector light={0,0,1.0};

static ZObject *ZObjectList;

void RenderScreen(Object *objs,matrix InvCamera)
{
  ZObject *ZObj,*ztemp;

  TransformObjects(objs,InvCamera);
  objs=CullObjects(objs);
  ZObjectList=0;
  RenderObject(objs);

  ZObjectList=ZSortList(ZObjectList);

  ZObj=ZObjectList;
  while(ZObj) {
    RenderZObject(ZObj);
    ztemp=ZObj;
    ZObj=ZObj->next;
    switch (ztemp->ObjType) {
    case TYPE_ZFLATPOLY:
	MemLeave(ztemp,MEM_ZFLATPOLY);
	break;
    case TYPE_ZTEXTUREPOLY:
	MemLeave(ztemp,MEM_ZTEXTUREPOLY);
	break;
    case TYPE_ZGOROUDPOLY:
	MemLeave(ztemp,MEM_ZGOROUDPOLY);
	break;
    }
  }


}

void RenderObject(Object *objs)
{
  int i;

  if (objs==0)
    return;

  switch(objs->Type) {
  case OBJ_CONTAINER:
    for (i=0;i<objs->U.Container.ModifyNumObjects;i++) {
      RenderObject(objs->U.Container.ModifyObject[i]);
    }
    break;
  case OBJ_PRIMITIVE:
    RenderPrimitives(objs->U.Primitive.Desc,objs->WorldTransform);
    break;
  }
    

  RenderObject(objs->modifynext);
}


void TransformObjects(Object *objs,matrix Transform)
{
  int i;

  while (objs) {
    MMult(Transform,objs->LocalTransform,objs->WorldTransform);
    if (objs->Type==OBJ_CONTAINER) {
      for (i=0;i<objs->U.Container.NumObjects;i++) {
	TransformObjects(objs->U.Container.Object[i],objs->WorldTransform);
      }
    }
    objs=objs->next;
  }
}

Object *CullObjects(Object *objs) 
{
  int i,numobj;
  real maxx,maxy,maxz,minx,miny,minz,r;

  if (objs==0)
    return 0;
  
  objs->modifynext=CullObjects(objs->next);
  

  switch(objs->Type) {
  case OBJ_PRIMITIVE:
    r=objs->U.Primitive.Desc->Radius;
    maxx=objs->WorldTransform[0][3]+r;
    minx=objs->WorldTransform[0][3]-r;
    maxy=objs->WorldTransform[1][3]+r;
    miny=objs->WorldTransform[1][3]-r;
    maxz=objs->WorldTransform[2][3]+r;
    minz=objs->WorldTransform[2][3]-r;
    if (minz>FarPlane)
      return 0;
    if (maxz<NearPlane)
      return 0;
    if (maxx<-maxz)
      return 0;
    if (minx>maxz)
      return 0;
    if (maxy<-maxz)
      return 0;
    if (miny>maxz)
      return 0;

    return objs;
    /* break;*/
  case OBJ_CONTAINER:
    numobj=0;
    for (i=0;i<objs->U.Container.NumObjects;i++) {
      if ((objs->U.Container.ModifyObject[numobj]=
	   CullObjects(objs->U.Container.Object[i]))!=0) {
	numobj++;
      }
      objs->U.Container.ModifyNumObjects=numobj;
    }
    return objs;
    /*break;*/
  }


  return 0;

}

#define Xmax 100
#define D 100

#define DATADIM 8

/* Data for object pipeline: */
vector EyePoint[MAXVERTICES];
signed_32bit ScreenPoint[MAXVERTICES*2];
signed_32bit VertexInt[MAXVERTICES];
byte PolyVisible[MAXPOLYGONS];
byte VertexVisible[MAXVERTICES];


void RenderPrimitives(ObjectDesc *desc,matrix transform)
{
  Polygon *poly;
  matrix world2obj;
  vector v;
  vector objlight;
  signed_32bit midz;
  int polynr,vertnr;
  int i,j;
  real dz1,dz2,dz1dz2;
  int from,to,tmp;
  signed_32bit Inten;
  byte visible;
  ZFlatPoly *zflat;
  ZTexturePoly *ztexture;
  ZGoroudPoly *zgoroud;
  int isback,numdata;
  signed_32bit tmpdata;
  static signed_32bit data[DATADIM*MAXVERTINPOLY]; /* static =>save stack*/

  MCopy(transform,world2obj);
  InvertM(world2obj);

  for (vertnr=0;vertnr<desc->NumVertices;vertnr++) {
    VertexVisible[vertnr]=0;
  }

  for (polynr=0;polynr<desc->NumPolys;polynr++) {
    poly=&desc->Poly[polynr];
    v.x=desc->Vertex[poly->Vertex[0]].x-world2obj[0][3];
    v.y=desc->Vertex[poly->Vertex[0]].y-world2obj[1][3];
    v.z=desc->Vertex[poly->Vertex[0]].z-world2obj[2][3];
    PolyVisible[polynr]=visible=(VDot(&poly->Normal,&v) <= 0);

    if (visible) {
      if (poly->Type==POLY_GOROUD) {
	for (i=0;i<poly->NumVertices;i++) {
	  VertexVisible[poly->Vertex[i]]=0x3; /* Visible and shade */
	}
      } else {
	for (i=0;i<poly->NumVertices;i++) {
	  VertexVisible[poly->Vertex[i]]|=0x1; /* Visible at least */
	}
      }
    }
  }

  /* Calculate the light vector in object space: */
  VMult3(world2obj,&light,&objlight);

  for (vertnr=0;vertnr<desc->NumVertices;vertnr++) {
    if (VertexVisible[vertnr]) {
      VMult4(transform,&desc->Vertex[vertnr],&EyePoint[vertnr]);
      if (EyePoint[vertnr].z>NearPlane) {
      ScreenPoint[vertnr*2]=
	8*HW_ScreenSizeX*D/Xmax*EyePoint[vertnr].x/EyePoint[vertnr].z
	+ HW_ScreenCenterX;
      ScreenPoint[vertnr*2+1]=
	8*HW_ScreenSizeX*D/Xmax*EyePoint[vertnr].y/EyePoint[vertnr].z
	+ HW_ScreenCenterY;
      }
    }
    if (VertexVisible[vertnr]&0x2) {
      Inten=-VDot(&objlight,&desc->VertexNormal[vertnr])*65536;
      if (Inten<100)
	Inten=100;
      if (Inten>=65536-100)
	Inten=65536-100;
      VertexInt[vertnr]=Inten;
    }
  }


/* Generate ZObjects and add them to ZObjectList */
/* Splice polys to triangles, and in the process clip them to z>=NearPlane */
  for (polynr=0;polynr<desc->NumPolys;polynr++) {
    if (PolyVisible[polynr]) {
      poly=&desc->Poly[polynr];
      
      numdata=0;
      isback=EyePoint[poly->Vertex[0]].z>=NearPlane;
      for (i=0;i<poly->NumVertices;i++) {
	if (isback) {
	  data[numdata*DATADIM+0]=ScreenPoint[poly->Vertex[i]*2];
	  data[numdata*DATADIM+1]=ScreenPoint[poly->Vertex[i]*2+1];
	  data[numdata*DATADIM+2]=EyePoint[poly->Vertex[i]].z;
	  if (poly->Type==POLY_GOROUD) {
	    data[numdata*DATADIM+3]=VertexInt[poly->Vertex[i]];
	  } else if (poly->Type==POLY_TEXTURE) {
	    tmpdata=poly->UV[i*2]*(1<<16);
	    if (tmpdata<500)
	      tmpdata=500;
	    data[numdata*DATADIM+3]=tmpdata;  
	    tmpdata=poly->UV[i*2+1]*(1<<16);
	    if (tmpdata<500)
	      tmpdata=500;
	    data[numdata*DATADIM+4]=tmpdata;
	  }
	  numdata++;
	}
	from=i;
	to=i+1;
	if (to>=poly->NumVertices)
	  to=0;
	if ((isback && (EyePoint[poly->Vertex[to]].z<NearPlane)) ||
	    (!isback && (EyePoint[poly->Vertex[to]].z>=NearPlane)) ) {
	  /* Switched side of the clipping plane */
	  isback=!isback;
	  if (EyePoint[poly->Vertex[from]].z >
	      EyePoint[poly->Vertex[to]].z) {
	    /* Always clip from far to near => 
	       => Always clips same line same way => same coordinate */
	    tmp=from;
	    from=to;
	    to=tmp;
	  }
	  data[numdata*DATADIM+2]=NearPlane;

	  dz1=NearPlane-EyePoint[poly->Vertex[from]].z;
	  dz2=EyePoint[poly->Vertex[to]].z-EyePoint[poly->Vertex[from]].z;
	  dz1dz2=dz1/dz2;
	  data[numdata*DATADIM+0]=
	    8*HW_ScreenSizeX*D/Xmax*(
	      (EyePoint[poly->Vertex[to]].x - 
	       EyePoint[poly->Vertex[from]].x)*dz1dz2 
	      + EyePoint[poly->Vertex[from]].x)/NearPlane
	    + HW_ScreenCenterX;
	  data[numdata*DATADIM+1]=
	    8*HW_ScreenSizeX*D/Xmax*(
	      (EyePoint[poly->Vertex[to]].y - 
	       EyePoint[poly->Vertex[from]].y)*dz1dz2 
	      + EyePoint[poly->Vertex[from]].y)/NearPlane
	    + HW_ScreenCenterY;
	  if (poly->Type==POLY_GOROUD) {
	    data[numdata*DATADIM+3]=
	      (VertexInt[poly->Vertex[to]] -
	       VertexInt[poly->Vertex[from]])*dz1dz2
	      + VertexInt[poly->Vertex[from]];
	  } else if (poly->Type==POLY_TEXTURE) {
	    data[numdata*DATADIM+3]=
	      ((poly->UV[to*2]-poly->UV[from*2])*dz1dz2
	       + poly->UV[from*2]) * (1<<16);
	    data[numdata*DATADIM+4]=
	      ((poly->UV[to*2+1]-poly->UV[from*2+1])*dz1dz2
	       + poly->UV[from*2+1])*(1<<16);
	  }
	  numdata++;
	}
      }

      switch(poly->Type) {
      case POLY_FLAT:
	zflat=MemGet(MEM_ZFLATPOLY);
	assert(zflat!=0);
	zflat->ObjType=TYPE_ZFLATPOLY;
	zflat->ClockWise=1;
	zflat->Color=poly->Color;
	zflat->Intensity=-VDot(&poly->Normal,&objlight)*65536;
	if (zflat->Intensity>=65536-100)
	  zflat->Intensity=65535-100;
	if (zflat->Intensity<100)
	  zflat->Intensity=100;
	
	for (j=1;j<=numdata-2;j++) {
	  midz=0;
	  zflat->Vertices[0]=data[0];
	  zflat->Vertices[1]=data[1];	
	  midz+=data[2];
	  zflat->Vertices[1*2]=data[j*DATADIM];
	  zflat->Vertices[1*2+1]=data[j*DATADIM+1];
	  midz+=data[j*DATADIM+2];
	  zflat->Vertices[2*2]=data[(j+1)*DATADIM];
	  zflat->Vertices[2*2+1]=data[(j+1)*DATADIM+1];
	  midz+=data[(j+1)*DATADIM+2];
	  
	  zflat->ZValue=midz;
	  zflat->next=ZObjectList;
	  ZObjectList=(ZObject *)zflat;
	  
	  zflat=MemGet(MEM_ZFLATPOLY);
	  assert(zflat!=0);
	  zflat->ObjType=TYPE_ZFLATPOLY;
	  zflat->ClockWise=1;
	  zflat->Color=((ZFlatPoly *)ZObjectList)->Color;
	  zflat->Intensity=((ZFlatPoly *)ZObjectList)->Intensity;
	}
 	MemLeave(zflat,MEM_ZFLATPOLY);
	break;
      case POLY_GOROUD:
	zgoroud=MemGet(MEM_ZGOROUDPOLY);
	assert(zgoroud!=0);
	zgoroud->ObjType=TYPE_ZGOROUDPOLY;
	zgoroud->ClockWise=1;
	zgoroud->Color=poly->Color;
	for (j=1;j<=numdata-2;j++) {
	  midz=0;
	  zgoroud->Vertices[0]=data[0];
	  zgoroud->Vertices[1]=data[1];	
	  zgoroud->Vertices[2]=data[3];
	  midz+=data[2];
	  zgoroud->Vertices[1*3]=data[j*DATADIM];
	  zgoroud->Vertices[1*3+1]=data[j*DATADIM+1];
	  zgoroud->Vertices[1*3+2]=data[j*DATADIM+3];	
	  midz+=data[j*DATADIM+2];
	  zgoroud->Vertices[2*3]=data[(j+1)*DATADIM];
	  zgoroud->Vertices[2*3+1]=data[(j+1)*DATADIM+1];
	  zgoroud->Vertices[2*3+2]=data[(j+1)*DATADIM+3];	
	  midz+=data[(j+1)*DATADIM+2];
	  
	  zgoroud->ZValue=midz;
	  zgoroud->next=ZObjectList;
	  ZObjectList=(ZObject *)zgoroud;
	  
	  zgoroud=MemGet(MEM_ZGOROUDPOLY);
	  assert(zgoroud!=0);
	  zgoroud->ObjType=TYPE_ZGOROUDPOLY;
	  zgoroud->ClockWise=1;
	  zgoroud->Color=poly->Color;
	}
 	MemLeave(zgoroud,MEM_ZGOROUDPOLY);
	break;
      case POLY_TEXTURE:
	ztexture=MemGet(MEM_ZTEXTUREPOLY);
	assert(ztexture!=0);
	ztexture->ObjType=TYPE_ZTEXTUREPOLY;
	ztexture->ClockWise=1;
	ztexture->BitWidth=poly->BitWidth;
	ztexture->Texture=poly->Texture;
	for (j=1;j<=numdata-2;j++) {
	  midz=0;
	  ztexture->Vertices[0]=data[0];
	  ztexture->Vertices[1]=data[1];	
	  ztexture->Vertices[2]=data[3];
	  ztexture->Vertices[3]=data[4];
	  midz+=data[2];
	  ztexture->Vertices[1*4]=data[j*DATADIM];
	  ztexture->Vertices[1*4+1]=data[j*DATADIM+1];
	  ztexture->Vertices[1*4+2]=data[j*DATADIM+3];	
	  ztexture->Vertices[1*4+3]=data[j*DATADIM+4];	
	  midz+=data[j*DATADIM+2];
	  ztexture->Vertices[2*4]=data[(j+1)*DATADIM];
	  ztexture->Vertices[2*4+1]=data[(j+1)*DATADIM+1];
	  ztexture->Vertices[2*4+2]=data[(j+1)*DATADIM+3];	
	  ztexture->Vertices[2*4+3]=data[(j+1)*DATADIM+4];	
	  midz+=data[(j+1)*DATADIM+2];

	  ztexture->ZValue=midz;
	  ztexture->next=ZObjectList;
	  ZObjectList=(ZObject *)ztexture;
	  
	  ztexture=MemGet(MEM_ZTEXTUREPOLY);
	  assert(ztexture!=0);
	  ztexture->ObjType=TYPE_ZTEXTUREPOLY;
	  ztexture->ClockWise=1;
	  ztexture->BitWidth=poly->BitWidth;
	  ztexture->Texture=poly->Texture;
	}
 	MemLeave(ztexture,MEM_ZTEXTUREPOLY);
	break;
      }
    }
  }
}



