#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>

#ifdef MSDOS
#include <pc.h>
#include <dos.h>
#endif

#include <arch.h>
#include <device.h>
#include <common.h>
#include <dbgmem.h>
#include <error.h>
#include "vector.h"
#include "zobject.h"
#include "object.h"
#include "engine.h"
#include "render.h"
#include "zsort.h"
#include "readobj.h"

void usage()
{
  printf("Usage: show filename [-f] [-p] [-a] [-x xsize] [-y ysize]\n"
	 "Files must be .gem or .geo\n"
	 "Flags: -f Flat shading, -p Page Flipping, -a accelarated clearing (when page flipping)\n"
	 "       -x & -y resolution\n"
	 "Default resolution is 640 x 480, and is always 8 bit\n"
	 "Note: page flipping is faster under X target, but will flicker a lot under KGI due to no support for vblank wait.\n"
	 "Use arrow keys to rotate object, space to turn on/off rotation. Read docs for other keys.\n"
	 );
}
  
int main(int argc, char *argv[])
{
  Object obj;
  matrix m=UNITMATRIX;
  matrix rot1,rot2,rot3,tmp;
  matrix rot1m,rot2m,rot3m;
  matrix camera=UNITMATRIX;
  vector tmpvect;
  byte *pal;
  byte *parrot;
  int fd;
  int quit=0;
  int rotate=1;
  int rotatelight=0;
  int c;
  signed_32bit z;
  FILE *fil;
  int frame=0;
  int polygon;
  int color;
  int flash=0;

  double elapsed;
  struct timeval t0, t1;

  /* options: */
  int shade=1;
  int x=640,y=480;
  int page_flip=0;
  int accel_clear=0;
  char *filename;
  
  while (1) {
    int c;
    c =  getopt(argc,argv,"fpax:y:");
    if (c==-1)
      break;
    switch(c) {
    case 'f':
      shade = 0;
      break;
    case 'p':
      page_flip = 1;
      break;
    case 'a':
      accel_clear = 1;
      break;
    case 'x':
      x = atoi(optarg);
      break;
    case 'y':
      y = atoi(optarg);
      break;
    case '?':
      usage();
      return 1;
      /*break;*/
    }
  }

#ifdef DEVICE_LIBGGI
  HW_libggi_setvars(x,y,page_flip,accel_clear);
#endif

  
  if ( (argc - optind) != 1 ) {
    usage();
    return 1;
  }
  filename = argv[optind];
    
  parrot=malloc(4096);
  fd=open("parrot.raw",OPEN_READ_BINARY);
  read(fd,parrot,4096);
  close(fd);

  MRotZ(3/2.0*1.745329e-2,rot1);
  MRotX(5/2.0*1.745329e-2,rot2);
  MRotY(7/2.0*1.745329e-2,rot3);
  MRotZ(-3/2.0*1.745329e-2,rot1m);
  MRotX(-5/2.0*1.745329e-2,rot2m);
  MRotY(-7/2.0*1.745329e-2,rot3m);

  if (!InitRenderer()) {
    printf("Couldn't init renderer\n");
    exit(1);
  }

  if (!HW_OpenScreen())
    exit(1);

  pal=malloc(256*3);
  fd=open("standard.pal",OPEN_READ_BINARY);
  read(fd,pal,256*3);
  close(fd);
  HW_SetPalette(pal);
  free(pal);

  obj.Type=OBJ_PRIMITIVE;
  obj.next=0;
  MCopy(m,obj.LocalTransform);
  fil=fopen(filename,"rt");
  assert(fil!=0);
  if (strcmp(&filename[strlen(filename)-4],".gem")==0)
    obj.U.Primitive.Desc=ReadGEM(fil,1200.0,shade);
  else if (strcmp(&filename[strlen(filename)-4],".geo")==0)
    obj.U.Primitive.Desc=ReadGEO(fil,1200.0,shade);
  else {
    printf("Unknown file type!\n");
    exit(1);
  }

  polygon=0;
  color=obj.U.Primitive.Desc->Poly[polygon].Color;
  flash=0;

  gettimeofday(&t0, NULL);

  z=2000;
#ifdef HW_HAS_KEYBOARD
  while (!quit) {
#else
  while (frame<1000) {
#endif
    frame++;
    camera[2][3]=z;
    /*HW_MemSet(HW_OffscreenBuffer,0,HW_BytesPerRow*HW_ScreenSizeY);*/
    HW_ClearOffscreen();

    RenderScreen(&obj,camera);
    HW_DisplayScreen();
    if (rotate) {
      MMult(camera,rot1,tmp);
      MMult(tmp,rot2,camera);
      MMult(camera,rot3,tmp);
      MCopy(tmp,camera);
      MOrtoNormalize(camera);
    }
    if (rotatelight) {
      VMult3(rot3m,&light,&tmpvect);
      light = tmpvect;
    }
    
#ifdef HW_HAS_KEYBOARD
    while(HW_KeyboardHit()) {
      c=HW_GetKey();
      switch(c) {
      case 27:
	quit=1;
	break;
      case 'l':
	rotatelight = !rotatelight;
	break;
      case 'f':
	obj.U.Primitive.Desc->Poly[polygon].Type=POLY_FLAT;
	break;  
      case 'g':
	obj.U.Primitive.Desc->Poly[polygon].Type=POLY_GOROUD;
	break;  
      case 't':
	obj.U.Primitive.Desc->Poly[polygon].Type=POLY_TEXTURE;
	obj.U.Primitive.Desc->Poly[polygon].Texture=parrot;
	obj.U.Primitive.Desc->Poly[polygon].BitWidth=6;
	obj.U.Primitive.Desc->Poly[polygon].UV[0]=0;
	obj.U.Primitive.Desc->Poly[polygon].UV[1]=0;
	obj.U.Primitive.Desc->Poly[polygon].UV[2]=63;
	obj.U.Primitive.Desc->Poly[polygon].UV[3]=0;
	obj.U.Primitive.Desc->Poly[polygon].UV[4]=63;
	obj.U.Primitive.Desc->Poly[polygon].UV[5]=63;
	obj.U.Primitive.Desc->Poly[polygon].UV[6]=0;
	obj.U.Primitive.Desc->Poly[polygon].UV[7]=63;
	break;
     case '+':
	obj.U.Primitive.Desc->Poly[polygon].Color=color;
	polygon=(polygon+1)%obj.U.Primitive.Desc->NumPolys;
	color=obj.U.Primitive.Desc->Poly[polygon].Color;
	break;
      case '-':
	obj.U.Primitive.Desc->Poly[polygon].Color=color;
	polygon=(polygon-1+obj.U.Primitive.Desc->NumPolys)%obj.U.Primitive.Desc->NumPolys;
	color=obj.U.Primitive.Desc->Poly[polygon].Color;
	break;
      case HW_KEY_KP7:
      case '7': z-=100; break;
      case HW_KEY_KP1:
      case '1': z+=100; break;
      case HW_KEY_KP8:
      case '8': camera[1][3]-=100; break;
      case HW_KEY_KP2:
      case '2': camera[1][3]+=100; break;
      case HW_KEY_KP4:
      case '4': camera[0][3]-=100; break;
      case HW_KEY_KP6:
      case '6': camera[0][3]+=100; break;
      case 'c':
	color=(color+1)%16;
	break;
      case HW_KEY_HOME:
	MMult(camera,rot1,tmp);
	MCopy(tmp,camera);
	break;
      case HW_KEY_END:
	MMult(camera,rot1m,tmp);
	MCopy(tmp,camera);
	break;
      case HW_KEY_UP:
	MMult(camera,rot2,tmp);
	MCopy(tmp,camera);
	break;
      case HW_KEY_DOWN:
	MMult(camera,rot2m,tmp);
	MCopy(tmp,camera);
	break;
      case HW_KEY_LEFT:
	MMult(camera,rot3,tmp);
	MCopy(tmp,camera);
	break;
      case HW_KEY_RIGHT:
	MMult(camera,rot3m,tmp);
	MCopy(tmp,camera);
	break;
     case ' ':
	rotate=!rotate;
	break;
      }
    }
    flash++;
    if (flash==5) {
      flash=0;
      if (color==0) 
	obj.U.Primitive.Desc->Poly[polygon].Color=1;
      else 
	obj.U.Primitive.Desc->Poly[polygon].Color=0;
    } else {
	obj.U.Primitive.Desc->Poly[polygon].Color=color;
    }
#endif /* HW_HAS_KEYBOARD */
  }
  gettimeofday(&t1, NULL);
  elapsed = (t1.tv_sec-t0.tv_sec)+(t1.tv_usec-t0.tv_usec)/1000000.0;
  
  HW_CloseScreen();
  free(parrot);

  printf("Frames per second: %f\n",frame/elapsed);

#ifdef MEM_DEBUG
  DbgMemStat();
#endif
  return 0;
}


